@@ -6,6 +6,8 @@
 from asn1.asn1 import *
 from asn1.ber import *
 from volatile import *
+from base_classes import BasePacket
+
 
 #####################
 #### ASN1 Fields ####
@@ -76,7 +78,7 @@ class ASN1F_field(ASN1F_element):
         if type(x) is list:
             x = x[:]
             for i in xrange(len(x)):
-                if isinstance(x[i], Packet):
+                if isinstance(x[i], BasePacket):
                     x[i] = x[i].copy()
         return x
 
@@ -113,9 +115,17 @@ class ASN1F_INTEGER(ASN1F_field):
     def randval(self):
         return RandNum(-2**64, 2**64-1)
 
+class ASN1F_BOOLEAN(ASN1F_field):
+    ASN1_tag= ASN1_Class_UNIVERSAL.BOOLEAN
+    def randval(self):
+        return RandChoice(True,False)
+
 class ASN1F_NULL(ASN1F_INTEGER):
     ASN1_tag= ASN1_Class_UNIVERSAL.NULL
 
+class ASN1F_SEP(ASN1F_NULL):
+    ASN1_tag= ASN1_Class_UNIVERSAL.SEP
+
 class ASN1F_enum_INTEGER(ASN1F_INTEGER):
     def __init__(self, name, default, enum):
         ASN1F_INTEGER.__init__(self, name, default)
@@ -148,6 +158,9 @@ class ASN1F_enum_INTEGER(ASN1F_INTEGER):
         else:
             return self.i2repr_one(pkt,x)
 
+class ASN1F_ENUMERATED(ASN1F_enum_INTEGER):
+    ASN1_tag = ASN1_Class_UNIVERSAL.ENUMERATED
+
 class ASN1F_STRING(ASN1F_field):
     ASN1_tag = ASN1_Class_UNIVERSAL.STRING
     def randval(self):
@@ -168,6 +181,9 @@ class ASN1F_TIME_TICKS(ASN1F_INTEGER):
 class ASN1F_UTC_TIME(ASN1F_STRING):
     ASN1_tag = ASN1_Class_UNIVERSAL.UTC_TIME
 
+class ASN1F_GENERALIZED_TIME(ASN1F_STRING):
+    ASN1_tag = ASN1_Class_UNIVERSAL.GENERALIZED_TIME
+
 class ASN1F_OID(ASN1F_field):
     ASN1_tag = ASN1_Class_UNIVERSAL.OID
     def randval(self):
@@ -273,7 +289,7 @@ class ASN1F_PACKET(ASN1F_field):
             c = cls(x)
         except ASN1F_badsequence:
             c = packet.Raw(x)
-        cpad = c[packet.Padding]
+        cpad = c.getlayer(packet.Padding)
         x = ""
         if cpad is not None:
             x = cpad.load
